"""Module holding metadata."""

from importlib.metadata import Distribution

from .i18n import gettext

_DISTRIBUTION = Distribution.from_name(
    "deskmaid",
)
_METADATA = _DISTRIBUTION.metadata

__author__ = _METADATA["Author"]
__email__ = _METADATA["Author-email"]
__license__ = _METADATA["License"]
__version__ = _METADATA["Version"]
__maintainer__ = _METADATA["Maintainer"]
__summary__ = gettext("Configurable desktop cleaner.")
__copyright__ = f"{__author__} <{__email__}>"
__issues__ = "https://github.com/Dashstrom/deskmaid/issues"



"""Utils directories."""

import sys
from pathlib import Path, PurePath
from typing import Tuple, Union

from .i18n import gettext


def get_root_directory() -> Path:
    """Get root directory."""
    if hasattr(sys, "_MEIPASS"):
        return Path(sys._MEIPASS)  # noqa: SLF001
    if getattr(sys, "frozen", False):
        return Path(sys.executable).parent
    return Path(__file__).parent


def get_config_directory() -> Path:
    """Get config directory."""
    return Path.home() / ".config" / "deskmaid"


PathLike = Union[str, PurePath]


def to_path(path: PathLike) -> Path:
    """Get path."""
    if isinstance(path, Path):
        return path
    if isinstance(path, PurePath):
        return Path(path)
    if isinstance(path, str):
        return Path(path)
    raise TypeError(gettext("%(path)s is not a valid path") % {"path": path})


EXTRA_EXT = ["gz", "bz2", "xz"]


def split_filename(path: Path) -> Tuple[str, str]:
    """Split name and suffix."""
    # Split extension and fix it for empty and compression
    name = path.stem
    ext = path.suffix

    # When file has two extension and one is an extra
    if ext.split(".")[-1] in EXTRA_EXT and "." in name:
        name, ext = name.rsplit(".", 1)
        ext = "." + ext + path.suffix

    # If file is like .<ext> without something before the dot
    # Use name as ext
    if not ext and "." in name:
        ext = name
        name = ""

    # Return the name and the casefold extension
    return name, ext.casefold()



"""Build package."""

from pathlib import Path

import babel.messages.frontend

ROOT_DIR = Path(__file__).parent
PO_DIR = ROOT_DIR / "deskmaid" / "locale"


def create_mo_files() -> None:
    cli = babel.messages.frontend.CommandLineInterface()
    cli.run(["pybabel", "compile", "--directory", str(PO_DIR)])  # type: ignore[no-untyped-call]


if __name__ == "__main__":
    create_mo_files()



"""Locales."""

import gettext as _gettext
import locale
import os
import pathlib
from typing import Union

Translation = Union[_gettext.GNUTranslations, _gettext.NullTranslations]


def load_translations() -> Translation:
    """Load translations."""
    locale_path = str(pathlib.Path(__file__).parent / "locale")
    try:
        translation: Translation = _gettext.translation(
            "messages", localedir=locale_path
        )
    except FileNotFoundError:
        if os.name == "nt":
            import ctypes

            windll = ctypes.windll.kernel32
            lang = locale.windows_locale[windll.GetUserDefaultUILanguage()]
            os.environ["LANG"] = lang
        translation = _gettext.translation(
            "messages",
            localedir=locale_path,
            fallback=True,
        )
    translation.install()
    _gettext.bindtextdomain("messages", localedir=locale_path)
    _gettext.textdomain("messages")
    return translation


_translation = load_translations()
gettext = _translation.gettext
ngettext = _translation.ngettext



"""Config for deskmaid."""

import logging
import pathlib
import re
import shutil
from pathlib import Path
from typing import Any, Dict, Match, Optional, Set, Type

from msgspec import Struct, field, json, yaml

from .file import get_config_directory, get_root_directory
from .i18n import gettext

logger = logging.getLogger(__name__)


class TemplateString:
    __slots__ = ("_template",)

    _RE_PATTERN = re.compile(
        r"(?P<before>[^\{]|)\{(?P<variable>[^\}]+)\}(?P<after>[^\}]|)"
    )

    def __init__(self, template: str) -> None:
        """Template for str."""
        self._template = template

    def render(self, **variables: Any) -> str:
        """Render the template with provided variables."""

        def repl(match: Match[str]) -> str:
            var = match["variable"]
            return (
                match["before"] + str(variables.get(var, var)) + match["after"]
            )

        return self._RE_PATTERN.sub(repl, self._template)


class TemplatePath:
    __slots__ = ("_template",)

    def __init__(self, template: str) -> None:
        """Template for path."""
        self._template = TemplateString(template)

    def render(self, **variables: Any) -> Path:
        """Render the template with provided variables."""
        return Path(
            self._template.render(
                root=get_root_directory(),
                home=Path.home(),
                **variables,
            )
        )


class FilterConfig(Struct, kw_only=True, forbid_unknown_fields=True):
    extensions: Set[str] = field(default_factory=set)
    files: Set[str] = field(default_factory=set)
    headers: Set[str] = field(default_factory=set)


def dec_hook(type: Type[Any], value: Any) -> Any:  # noqa: A002
    """Decode hook for msgspec."""
    if type is TemplateString:
        return TemplateString(value)
    if type is TemplatePath:
        return TemplatePath(value)
    if type is pathlib.Path:
        return pathlib.Path(value)
    error_message = gettext("Invalid type %(type)r for %(value)r") % {
        "type": type,
        "value": value,
    }
    raise ValueError(error_message)


def enc_hook(value: Any) -> Any:
    """Decode hook for msgspec."""
    if isinstance(value, TemplateString):
        return str(value)
    if isinstance(value, TemplatePath):
        return str(value)
    if isinstance(value, pathlib.Path):
        return str(value)
    error_message = gettext("Invalid type %(type)r for %(value)r") % {
        "type": type(value),
        "value": value,
    }
    raise ValueError(error_message)


class Move(Struct, kw_only=True, forbid_unknown_fields=True):
    src: Path
    dst: Path

    def do(self, dry_run: Optional[bool] = None) -> None:
        """Move the file."""
        logger.info(gettext("Move %s to %s"), self.src, self.dst)
        if not dry_run:
            self.dst.parent.mkdir(exist_ok=True, parents=True)
            shutil.move(str(self.src), str(self.dst))

    def undo(self, dry_run: Optional[bool] = None) -> None:
        """Undo the move."""
        logger.info(gettext("Undo %s to %s"), self.src, self.dst)
        if not dry_run:
            self.src.parent.mkdir(exist_ok=True, parents=True)
            shutil.move(str(self.dst), str(self.src))

    @staticmethod
    def load(line: bytes) -> "Move":
        """Load Move from bytes."""
        return json.decode(line, dec_hook=dec_hook, type=Move)

    def dump(self) -> bytes:
        """Dump Move into bytes."""
        return json.encode(self, enc_hook=enc_hook) + b"\n"


class Config(Struct, kw_only=True, forbid_unknown_fields=True):
    desktop: TemplatePath
    storage: TemplatePath
    filters: Dict[str, FilterConfig]
    fallback: str
    title: str
    app_icon: TemplatePath
    storage_icon: TemplatePath
    ignore: FilterConfig = field(default_factory=FilterConfig)
    history: TemplatePath

    @staticmethod
    def load(
        path: Optional[Path] = None,
        dry_run: Optional[bool] = None,
    ) -> "Config":
        """Load config from path or create it."""
        if path is None:
            if dry_run:
                path = get_root_directory() / "config" / "config.yml"
            else:
                path = get_config_directory() / "config.yml"
                if not path.exists():
                    template = get_root_directory() / "config" / "config.yml"
                    path.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(template, path)
        content = path.read_bytes()
        return yaml.decode(content, type=Config, dec_hook=dec_hook)



"""Module for command line interface."""

import argparse
import logging
import sys
from typing import NoReturn, Optional, Sequence

from .deskmaid import Deskmaid
from .i18n import gettext
from .info import __issues__, __summary__, __version__

LOG_LEVELS = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
logger = logging.getLogger(__name__)


class HelpArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:
        """Handle error from argparse.ArgumentParser."""
        self.print_help(sys.stderr)
        self.exit(2, f"{self.prog}: error: {message}\n")


def get_parser() -> argparse.ArgumentParser:
    """Prepare ArgumentParser."""
    parser = HelpArgumentParser(
        prog="deskmaid",
        description=__summary__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s, version {__version__}",
        help=gettext("show program's version number and exit"),
    )

    # Add subparsers
    subparsers = parser.add_subparsers(
        help=gettext("desired action to perform"),
        dest="action",
        required=True,
    )

    # Add parent parser with common arguments
    parent_parser = HelpArgumentParser(add_help=False)
    parent_parser.add_argument(
        "-v",
        "--verbose",
        help=gettext("verbose mode, enable INFO and DEBUG messages."),
        action="store_true",
        required=False,
    )

    # Parser of run command
    run_parser = subparsers.add_parser(
        "clean",
        parents=[parent_parser],
        help=gettext("run deskmaid for clean your desktop."),
    )
    run_parser.add_argument(
        "--dry-run",
        help=gettext("do not move files."),
        action="store_true",
        default=False,
        required=False,
    )
    run_parser.add_argument(
        "-r",
        "--recursive",
        help=gettext("recursive arrange."),
        action="store_true",
        default=False,
        required=False,
    )
    run_parser.add_argument(
        "--path",
        "-p",
        help=gettext("path to the directory to arrange."),
        required=False,
    )

    # Parser of undo command
    undo_parser = subparsers.add_parser(
        "undo",
        parents=[parent_parser],
        help=gettext("undo the last run command."),
    )
    undo_parser.add_argument(
        "--dry-run",
        help=gettext("do not move files."),
        action="store_true",
        default=False,
        required=False,
    )

    # Parser of shortcut command
    shortcut_parser = subparsers.add_parser(
        "shortcut",
        parents=[parent_parser],
        help=gettext("create shortcuts on desktop."),
    )
    shortcut_parser.add_argument(
        "--debug",
        help=gettext("show logs and wait before exit."),
        action="store_true",
        default=False,
        required=False,
    )
    return parser


def setup_logging(verbose: Optional[bool] = None) -> None:
    """Do setup logging."""
    # Setup logging
    logging.basicConfig(
        level=logging.DEBUG if verbose else logging.WARNING,
        format="[%(asctime)s] [%(levelname)s] [%(name)s] %(message)s",
    )


def entrypoint(argv: Optional[Sequence[str]] = None) -> None:
    """Entrypoint for command line interface."""
    try:
        parser = get_parser()
        args = parser.parse_args(argv)
        setup_logging(args.verbose)
        if args.action == "clean":
            maid = Deskmaid.load(dry_run=args.dry_run)
            maid.clean(
                path=args.path,
                dry_run=args.dry_run,
                recursive=args.recursive,
            )
        elif args.action == "undo":
            maid = Deskmaid.load(dry_run=args.dry_run)
            maid.undo(dry_run=args.dry_run)
        elif args.action == "shortcut":
            maid = Deskmaid.load()
            maid.shortcut(debug=args.debug)
        else:
            parser.error(gettext("No command specified"))
    except Exception as err:  # NoQA: BLE001
        logger.critical(gettext("Unexpected error"), exc_info=err)
        logger.critical(
            gettext("Please, report this error to %s."),
            __issues__,
        )
        sys.exit(1)



"""Script for arrange desktop.

Use `config.yml` for options
"""

import logging
import os
import shlex
import sys
from collections import defaultdict
from io import BytesIO
from pathlib import Path
from time import sleep
from typing import IO, Dict, Optional

from plyer import notification

from .config import Config, Move
from .file import PathLike, split_filename, to_path
from .i18n import gettext, ngettext

logger = logging.getLogger(__name__)


class Deskmaid:
    def __init__(self, config: Config) -> None:
        """Instanciate deskmaid."""
        self.config = config

    @staticmethod
    def load(
        path: Optional[Path] = None,
        dry_run: Optional[bool] = None,
    ) -> "Deskmaid":
        """Load config from path or default location."""
        return Deskmaid(Config.load(path, dry_run=dry_run))

    def organize(  # noqa: C901, PLR0912
        self,
        *,
        path: Optional[PathLike] = None,
        recursive: Optional[bool] = None,
        dry_run: Optional[bool] = None,
    ) -> Dict[Path, Path]:
        """Organize all non-ignored files and return the moved files."""
        if dry_run is None:
            dry_run = False
        if recursive is None:
            recursive = False
        if path is None:
            target = self.config.desktop.render()
        else:
            target = to_path(path)
        storage = self.config.storage.render()
        if dry_run:
            storage.mkdir(parents=True, exist_ok=True)
        # Make fast mapping
        mapping = defaultdict(
            lambda: self.config.fallback,
            {
                extension.casefold(): name
                for name, filter in self.config.filters.items()  # noqa: A001
                for extension in filter.extensions
            },
        )

        # Support recursive
        iterator = target.glob("**/*") if recursive else target.iterdir()

        # Save all changes
        if dry_run:
            stream: IO[bytes] = BytesIO()
        else:
            history = self.config.history.render()
            history.parent.mkdir(parents=True, exist_ok=True)
            stream = history.open("wb")

        # Create mapping
        ignore_extensions = {
            ext.casefold() for ext in self.config.ignore.extensions
        }

        # Find all files to move and there destinations
        files: Dict[Path, Path] = {}
        with stream:
            # Iterate files with smallest files first
            for src in sorted(iterator, key=lambda p: p.stat().st_size):
                # Keep only files
                if not src.is_file():
                    continue

                # Perform intelligent separation between stem and ext
                _, ext = split_filename(src)
                if ext.startswith("."):
                    ext = ext[1:]

                # Apply filters
                if src.name in self.config.ignore.files:
                    continue
                if ext in ignore_extensions:
                    continue

                # Find destination
                directory = storage / mapping[ext]
                dst = directory / src.name

                # Deskmaid has been run on himself
                if src == dst:
                    continue

                # Move file and save it into history
                move = Move(src=src, dst=self.rename(dst))
                move.do(dry_run=dry_run)
                if not dry_run:
                    stream.write(move.dump())
                files[src] = dst
        return files

    def rename(self, path: Path) -> Path:
        """Fix filenames."""
        # CHeck if fix is required
        if not path.exists():
            return path

        name, ext = split_filename(path)

        # Iterate until find correct file
        counter = 0
        while True:
            counter += 1
            result = path.parent / f"{name} ({counter}){ext}".strip()
            if not result.exists():
                return result

    def clean(
        self,
        *,
        path: Optional[PathLike] = None,
        recursive: Optional[bool] = None,
        dry_run: Optional[bool] = None,
    ) -> None:
        """Run deskmaid."""
        self.notify(
            gettext("At your command, senpai")
            + " (\u2044 \u2044•\u2044ω\u2044•\u2044 \u2044)"
        )
        try:
            arranged = self.organize(
                path=path,
                dry_run=dry_run,
                recursive=recursive,
            )
            if arranged:
                self.notify(
                    ngettext(
                        "one files arranged",
                        "%(num)d file arranged",
                        len(arranged),
                    )
                    % {"num": len(arranged)}
                    + " ଘ(੭ˊᵕˋ)੭* ੈ✩‧₊"
                )
            else:
                self.notify(
                    gettext(
                        "Oni-chan, everything is tidy\n"
                        "It's just the two of us now\n"
                    )
                    + "(\u2044 \u2044>\u2044 ▽ \u2044<\u2044 \u2044)"
                )
        except Exception as err:
            self.notify(str(err))
            logger.exception("An error was occurred", exc_info=err)

    def undo(self, dry_run: Optional[bool] = None) -> None:
        """Undo last arrange."""
        if dry_run is None:
            dry_run = False
        history = self.config.history.render()
        if history.exists():
            lines = history.read_bytes().strip().split(b"\n")
            for line in lines:
                move = Move.load(line)
                move.undo(dry_run=dry_run)
            history.unlink()

    def notify(self, message: str) -> bool:
        """Send desktop notification."""
        try:
            notification.notify(
                title=self.config.title,
                message=message,
                app_name=self.config.title,
                app_icon=str(self.config.app_icon.render()),
                ticker=self.config.title,
            )
            sleep(1)
        except NotImplementedError:
            logger.exception(
                gettext("Failed to send notification: %s"),
                message,
            )
            return False
        return True

    def _create_desktop_shortcut_windows(self, *, debug: bool) -> None:
        """Create shortcut on windows."""
        # Import API

        from win32com.client import CDispatch, Dispatch

        # Get desktop path
        desktop = self.config.desktop.render()

        # Load API
        shell = Dispatch("WScript.Shell")

        # Resolve path of executable
        python_exe = Path(sys.executable)
        pythonw_exe = python_exe.parent / "pythonw.exe"
        if pythonw_exe.exists() and not debug:
            python_exe = pythonw_exe

        # Resolve current module name
        module_name = vars(sys.modules[__name__])["__package__"]

        # Create program shortcut
        shortcut_path = str(desktop / f"{self.config.title}.lnk")
        lnk: CDispatch = shell.CreateShortCut(shortcut_path)
        lnk.TargetPath = str(python_exe)
        args = ["-m", module_name, "run"]
        if debug:
            args.insert(0, "-i")
            args.append("--dry-run")
            args.append("--verbose")
        lnk.Arguments = shlex.join(args)
        lnk.IconLocation = str(self.config.app_icon.render())
        lnk.save()

        # Create directory shortcut
        shortcut_path = str(desktop / "Files.lnk")
        Path(shortcut_path).unlink(missing_ok=True)
        lnk = shell.CreateShortCut(shortcut_path)
        lnk.TargetPath = str(self.config.storage.render())
        lnk.IconLocation = str(self.config.storage_icon.render())
        lnk.save()

    def shortcut(self, debug: Optional[bool] = None) -> None:
        """Create shortcut on desktop."""
        if debug is None:
            debug = False
        if os.name == "nt":
            self._create_desktop_shortcut_windows(debug=debug)
        else:
            raise NotImplementedError
